/************************************************************************/
/*                                                                      */
/*                      Connect several AVRs over one wire              */
/*              Author: Peter Dannegger                                 */
/*                                                                      */
/************************************************************************/
//                      Target: ATmega644
/*
Master:         ATmega644
Address:        none
Data pin:       PB3 = OC0A (transmit) = PCINT11 (receive)
KEY0:           PB4
KEY1:           PB5
KEY2:           PB6
LED0:           PC1
LED1:           PC2
LED2:           PC3

Slave 1:        ATmega48
Address:        1
Data pin:       PB3 = OC2A (transmit) = PCINT3 (receive)
KEY0:           PB0
KEY1:           PB1
LED0:           PC0

Slave 2:        ATtiny261
Address:        2
Data pin:       PB3 = OC1B (transmit) = PCINT11 (receive)
KEY0:           PB2
LED0:           PA6
LED1:           PA7

Data Protocol:
1 * T: o-Bit
2 * T: 1-Bit
4 * T: Sync-Bit
*/

#include <avr/interrupt.h>
#include <avr/power.h>
#include "mydefs.h"

#define F_CPU           8e6                     // 8MHz
#include<util/delay.h>
#define BIT_CLK         10e3                    // Bit time T = 10kHz
#define PRESCALER       64                      // F_CPU / 64
#define PRESCALER_SET   (1<<CS01 | 1<<CS00)
#define	BIT_TIME	(F_CPU / BIT_CLK / PRESCALER)
#define DATA_SIZE       2                       // 2 Byte
#define DATA_TIME       (1 + 1.5 * (DATA_SIZE * 16 + 8) / BIT_CLK * 1e3) // ms

#define	KEY0_PIN        SBIT( PINB,  PB4 )
#define	KEY1_PIN        SBIT( PINB,  PB5 )
#define	KEY2_PIN        SBIT( PINB,  PB6 )
#define	LED0            SBIT( PORTC, PC1 )
#define	LED0_DDR        SBIT( DDRC,  PC1 )
#define	LED1            SBIT( PORTC, PC2 )
#define	LED1_DDR        SBIT( DDRC,  PC2 )
#define	LED2            SBIT( PORTC, PC3 )
#define	LED2_DDR        SBIT( DDRC,  PC3 )
#define	DATA            SBIT( PORTB, PB3 )
#define	DATA_DDR        SBIT( DDRB,  PB3 )
#define	DATA_PIN        SBIT( PINB,  PB3 )


/****************************** Transmission ****************************/

uint8_t tx_buff[DATA_SIZE];
uint8_t tx_idx;


ISR( TIMER0_COMPA_vect )                        // Transmit interrupt
{
  uint8_t *dp;					// locals for faster access

  if( tx_idx < DATA_SIZE * 8 ){
    OCR0A = BIT_TIME * 1 - 0.5;                 // 0-bit: 1T
    dp = &tx_buff[tx_idx++ / 8];                // get byte address
    if( *dp & 0x80 )                            // msb first
      OCR0A = BIT_TIME * 2 - 0.5;               // 1-bit: 2T
    *dp <<= 1;                                  // shift bits
  }else{
    OCR0A = BIT_TIME * 4 - 0.5;                 // sync:  4T
    if( DATA_PIN )                              // after output = high
      TCCR0B = 0;                               // then stop T0
  }
}


void tx_data( uint8_t addr, uint8_t data )
{
  PCMSK1 = 0;    				// disable receive
  TCCR0B = 0;					// stop T0
  volatil(tx_idx) = 0;                              // point to first bit
  volatil(tx_buff[0]) = addr;
  volatil(tx_buff[1]) = data;
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  while( DATA_PIN == 0 );                       // wait until pin high
  TCNT0 = 0;
  OCR0A = BIT_TIME * 4 - 0.5;                   // sync: 4T
  TCCR0A = 1<<COM0A0 | 1<<COM0A1;               // set on compare
  TCCR0B = 1<<FOC0A;                            // set output high
  TCCR0A = 1<<COM0A0                            // toggle pin on compare
         | 1<<WGM01;                            // Mode 2: CTC
  DATA_DDR = 1;                                 // tx output on
  TIFR0 = 1<<OCF0A;				// clear earlier interrupts
  TIMSK0 = 1<<OCIE0A;				// enable tx interrupt
  TCCR0B = PRESCALER_SET;                       // start T0
  while( TCCR0B );				// untile tx done
}


/****************************** Receiving *******************************/

#define rx_idx  tx_idx
#define rx_buff tx_buff
uint8_t rx_timeout;


ISR( PCINT1_vect )                              // Receive interrupt
{
  uint8_t *dp;					// locals for faster access
  uint8_t i;                                	// locals for faster access

  TCNT0 = 0;
  if( rx_idx >= DATA_SIZE * 8 ){                // on start or stop bit
    rx_idx = 0;                                 // point to first bit
  }else{
    dp = &rx_buff[rx_idx / 8];                  // get byte address
    i = *dp << 1;                               // shift bits
    if( TIFR0 & 1<<OCF0A )                      // if pulse > 1.4T
      i++;                                      // msb first
    *dp = i;
    rx_idx++;                                   // point to next bit
  }
  TIFR0 = 1<<OCF0A;                             // clear 2T detection flag
}


ISR( TIMER0_COMPB_vect )                        // sync received
{
  if( rx_idx == DATA_SIZE * 8 )                 // receive successful
    PCMSK1 = 0;					// disable further receive
  rx_idx = 255;                                 // detect sync transition
  if( rx_timeout ){				// timeout answer from slave
    rx_timeout--;
    TCNT0 = 0;
  }
}


#define	rx_busy() PCMSK1                        // 0 = data received


void rx_init( void )
{
  TCCR0B = 0;					// Stop T0
  volatil(rx_timeout) = 10;			// 28 * T
  volatil(rx_idx) = 255;                        // detect sync transition
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  TCCR0A = 0;					// disconnect tx from timer
  TCNT0 = 0;
  OCR0A = BIT_TIME * 1.4 - 0.5;                 // detect 2T
  OCR0B = BIT_TIME * 2.8 - 0.5;                 // detect 4T
  TIFR0 = 1<<OCF0A | 1<<OCF0B;
  TIMSK0 = 1<<OCIE0B;                           // sync interrupt
  TCCR0B = PRESCALER_SET;                       // start T0
  PCIFR = 1<<PCIF1;                             // clear earlier interrupts
  PCICR = 1<<PCIE1;                             // pin change interrupt
  PCMSK1 = 1<<PCINT11;
}


int main( void )
{
  clock_prescale_set( clock_div_1 );
  PORTA = 0xFF;					// input with pullup
  PORTB = 0xFF;					// input with pullup
  PORTC = 0xFF;					// input with pullup
  PORTD = 0xFF;					// input with pullup
  LED0 = 0;
  LED0_DDR = 1;                                 // Output
  LED1 = 0;
  LED1_DDR = 1;                                 // Output
  LED2 = 0;
  LED2_DDR = 1;                                 // Output

  sei();
  for(;;){
    tx_data( 1, KEY0_PIN );                     // slave 1
    rx_init();                                  // start receive
    do{
      if( rx_busy() == 0 ){
        LED0 = volatil(rx_buff[1]) & 0x01;
        LED1 = !!(volatil(rx_buff[1]) & 0x02);
        break;
      }
    }while( volatil(rx_timeout) );
    tx_data( 2, KEY1_PIN | KEY2_PIN<<1 );       // slave 2
    rx_init();                                  // start receive
    do{
      if( rx_busy() == 0 ){
        LED2 = volatil(rx_buff[1]) & 0x01;
        break;
      }
    }while( volatil(rx_timeout) );
  }
}
