/************************************************************************/
/*									*/
/*			Connect several AVRs over one wire              */
/*		Author: Peter Dannegger					*/
/*									*/
/************************************************************************/
//                      Target: ATtiny261
/*
Master:         ATmega644
Address:        none
Data pin:       PB3 = OC0A (transmit) = PCINT11 (receive)
KEY0:           PB4
KEY1:           PB5
KEY2:           PB6
LED0:           PC1
LED1:           PC2
LED2:           PC3

Slave 1:        ATmega48
Address:        1
Data pin:       PB3 = OC2A (transmit) = PCINT3 (receive)
KEY0:           PB0
KEY1:           PB1
LED0:           PC0

Slave 2:        ATtiny261
Address:        2
Data pin:       PB3 = OC1B (transmit) = PCINT11 (receive)
KEY0:           PB2
LED0:           PA6
LED1:           PA7

Data Protocol:
1 * T: o-Bit
2 * T: 1-Bit
4 * T: Sync-Bit
*/

#include <avr/interrupt.h>
#include <avr/power.h>
#include "mydefs.h"

#define F_CPU           8e6                     // 8MHz
#define BIT_CLK         10e3                    // Bit time T = 10kHz
#define PRESCALER       64                      // F_CPU / 64
#define PRESCALER_SET	(1<<CS12 | 1<<CS11 | 1<<CS10)
#define	BIT_TIME	(F_CPU / BIT_CLK / PRESCALER)
#define DATA_SIZE       2                       // 2 Byte

#define	KEY0_PIN        SBIT( PINB,  PB2 )
#define	LED0            SBIT( PORTA, PA6 )
#define	LED0_DDR        SBIT( DDRA,  PA6 )
#define	LED1            SBIT( PORTA, PA7 )
#define	LED1_DDR        SBIT( DDRA,  PA7 )
#define	DATA            SBIT( PORTB, PB3 )
#define	DATA_DDR        SBIT( DDRB,  PB3 )
#define	DATA_PIN        SBIT( PINB,  PB3 )


/****************************** Transmission ****************************/

uint8_t tx_buff[DATA_SIZE];
uint8_t tx_idx;


ISR( TIMER1_COMPB_vect )                        // Transmit interrupt
{
  uint8_t *dp;					// locals for faster access

  if( tx_idx < DATA_SIZE * 8 ){
    OCR1C = BIT_TIME * 1 - 0.5;                 // 0-bit: 1T
    dp = &tx_buff[tx_idx++ / 8];                // get byte address
    if( *dp & 0x80 )                            // msb first
      OCR1C = BIT_TIME * 2 - 0.5;               // 1-bit: 2T
    *dp <<= 1;                                  // shift bits
  }else{
    OCR1C = BIT_TIME * 4 - 0.5;                 // sync:  4T
    if( DATA_PIN )                              // after output = high
      TCCR1B = 0;                               // then stop T1
  }
}


void tx_data( uint8_t addr, uint8_t data )
{
  PCMSK1 = 0;    				// disable receive
  TCCR1B = 0;					// stop T1
  volatil(tx_idx) = 0;                          // point to first bit
  volatil(tx_buff[0]) = addr;
  volatil(tx_buff[1]) = data;
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  while( DATA_PIN == 0 );                       // wait until pin high
  TIFR = 1<<OCF1B;                              // clear earlier interrupts
  TCNT1 = 0;
  OCR1B = 0;
  OCR1C = BIT_TIME * 4 - 0.5;                   // sync: 4T
  TCCR1A = 1<<COM1B1 | 1<<COM1B0                // set on compare
         | 1<<FOC1B;                            // set output high
  TCCR1A = 1<<COM1B0;                           // toggle pin on compare
  DATA_DDR = 1;                                 // tx output on
//  nop();  nop();                                // 2 cycle to clear flags !
  TIMSK = 1<<OCIE1B;                            // enable tx interrupt
  TCCR1B = PRESCALER_SET;                       // start T1
  while( TCCR1B );                              // until tx done
}


/****************************** Receiving *******************************/

#define rx_idx  tx_idx
#define rx_buff tx_buff


ISR( PCINT_vect )                               // Receive interrupt
{
  uint8_t *dp;					// locals for faster access
  uint8_t i;                                	// locals for faster access

  TCNT1 = 0;
  if( rx_idx >= DATA_SIZE * 8 ){                // on start or stop bit
    rx_idx = 0;                                 // point to first bit
  }else{
    dp = &rx_buff[rx_idx / 8];                  // get byte address
    i = *dp << 1;                               // shift bits
    if( TIFR & 1<<OCF1B )                       // if pulse > 1.4T
      i++;                                      // msb first
    *dp = i;
    rx_idx++;                                   // point to next bit
  }
  TIFR = 1<<OCF1B;                              // clear 2T detection flag
}


ISR( TIMER1_COMPA_vect )                        // sync received
{
  if( rx_idx == DATA_SIZE * 8 )                 // receive successful
    PCMSK1 = 0;					// disable further receive
  rx_idx = 255;                                 // ready for start transition
}


#define	rx_busy() PCMSK1                        // 0 = data received


void rx_init( void )
{
  TCCR1B = 0;					// Stop T0
  volatil(rx_idx) = 255;                        // detect sync transition
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  TCCR1A = 0;					// disconnect tx output
  TIFR = 1<<OCF1A | 1<<OCF1B;                   // clear earlier interrupts
  TCNT1 = 0;
  OCR1B = BIT_TIME * 1.4 - 0.5;                 // detect 2T
  OCR1A = BIT_TIME * 2.8 - 0.5;                 // detect 4T
//  nop();  nop();                                // 2 cycle to clear flags !
  TIMSK = 1<<OCIE1A;                            // sync interrupt
  TCCR1B = PRESCALER_SET;                       // start T1
  PCMSK0 = 0;                                   // 0xC8 after reset !
  PCMSK1 = 1<<PCINT11;
  GIFR = 1<<PCIF;                               // clear earlier interrupts
  GIMSK = 1<<PCIE0;                             // pin change interrupt (8..11)
}


int main( void )
{
  clock_prescale_set( clock_div_1 );
  PORTA = 0xFF;					// input with pullup
  PORTB = 0xFF;					// input with pullup
  LED0 = 0;
  LED0_DDR = 1;                                 // Output
  LED1 = 0;
  LED1_DDR = 1;                                 // Output

  sei();
  for(;;){
    rx_init();                                  // start receive
    while( rx_busy() );
    if( volatil(rx_buff[0]) == 2 ){			// if slave 2 addressed
      LED0 = volatil(rx_buff[1]) & 1;
      LED1 = !!(volatil(rx_buff[1]) & 2);
      tx_data( 2, KEY0_PIN );
    }
  }
}
